unit RiscDisk;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, RiscFp, Dialogs;

const
  BUFSBYTE = 512;
  BUFSWORD = BUFSBYTE DIV 4;

type
  diskstatetype = (diskCommand, diskRead, diskWrite, diskWriting);

  bufty   =   array[0..Pred(BUFSWORD)  ] OF uint32_t;
  buftyp2 =   array[0..Pred(BUFSWORD)+2] OF uint32_t;
  bytebuf = array[0..pred(BUFSBYTE)  ] OF uint8_t;

  TDisk = object
  private
    state     : diskstatetype;
    myfile    : THandle;
    rx_buf    : bufty;
    rx_idx    : longint;
    tx_buf    : buftyp2;
    tx_cnt    : longint;
    tx_idx    : longint;
    offset    : uint32_t;
    procedure   read_sector(var buffer : bufty);
    procedure   read_sector2(var buffer : buftyp2);
    procedure   write_sector(buffer : bufty);
  public
    sdcard    : Boolean;
    constructor Create(filename : string);
    destructor  Destroy;
    procedure   run_command;
    function    read_: uint32_t;
    procedure   write_(value: uint32_t);
  end;

implementation

constructor TDisk.Create(filename : string);
var
  buffer : bufty;
begin
  state := diskCommand;
  sdcard := False;
  buffer[0] := 0;
  if not FileExists(filename) then begin
    ShowMessage('Can''t open file : "'+filename+'"');
    exit;
  end;
  myfile := fileOpen(filename, fmOpenReadWrite);
  read_sector(buffer);
  if (buffer[0] = $9B1EA38D) then offset := $80002 else offset := 0;
  sdcard := True;
end;

destructor TDisk.Destroy;
begin
  fileClose(myfile);
end;

procedure TDisk.write_(value: uint32_t);
begin
  inc(tx_idx);

  case state of
    diskCommand:
      if ((byte(value)<>$FF) OR (rx_idx<>0)) then begin
        rx_buf[rx_idx]:= value;
        inc(rx_idx);
        if rx_idx = 6 then begin
          run_command;
          rx_idx:= 0;
        end;
      end;

    diskRead:
      if tx_idx = tx_cnt  then begin
        state:= diskCommand;
        tx_cnt:= 0;
        tx_idx:= 0;
      end;

    diskWrite: if value = 254 then state:= diskWriting;

    diskWriting:
      begin
        if rx_idx < BUFSWORD then rx_buf[rx_idx]:= value;
        inc(rx_idx);
        if rx_idx = BUFSWORD then write_sector(rx_buf);
        if rx_idx = 130  then begin
          tx_buf[0] := 5;
          tx_cnt := 1;
          tx_idx:= -1;
          rx_idx:= 0;
          state:= diskCommand;
        end;
      end;
  end;{case}
end;

function TDisk.read_: uint32_t;
begin
  if (tx_idx >= 0) and (tx_idx < tx_cnt) then
    Result := tx_buf[tx_idx]
  else
    Result := 255;
end;

procedure TDisk.run_command;
var
  cmd: uint32_t;
  arg: uint32_t;
begin
  cmd := rx_buf[0];
  arg := (rx_buf[1] shl 24) or (rx_buf[2] shl 16) or (rx_buf[3] shl 8) or rx_buf[4];

  case cmd of
    81:
        begin
          state:= diskRead;
          tx_buf[0] := 0;
          tx_buf[1] := 254;
          fileseek(myfile, (arg - offset) * BUFSBYTE, fsFromBeginning);
          read_sector2(tx_buf);
          tx_cnt:=  2 + BUFSWORD;
        end;
    88:
        begin
          state:= diskWrite;
          fileseek(myfile, (arg - offset) * BUFSBYTE, fsFromBeginning);
          tx_buf[0] := 0;
          tx_cnt := 1;
        end;
  else
    tx_buf[0] := 0;
    tx_cnt := 1;
  end;(*case*)

  tx_idx := -1;
end;

procedure TDisk.read_sector(var buffer : bufty);
var
  bytes: bytebuf;
  i : 0..pred(BUFSWORD);
  i2 : 0..pred(BUFSBYTE);
begin
  for i2 := 0 TO pred(BUFSBYTE) DO bytes[i2] := 0;
  fileread(myfile, bytes, BUFSBYTE);
  for i := 0 to Pred(BUFSWORD)  DO
    buffer[i]:= longword(bytes[i*4+0]) or (longword(bytes[i*4+1]) shl 8) or (longword(bytes[i*4+2]) shl 16) or (longword(bytes[i*4+3]) shl 24);
end;

procedure TDisk.read_sector2(var buffer : buftyp2);
var
  bytes : bytebuf;
  i  : 0..pred(BUFSWORD);
  i2  : 0..pred(BUFSBYTE);
begin
  for i2 := 0 to pred(BUFSBYTE) DO bytes[i2] := 0;
  fileread(myfile, bytes, BUFSBYTE);
  for i := 0 to Pred(BUFSWORD)  DO
    buffer[i+2]:= longword(bytes[i*4+0]) or (longword(bytes[i*4+1]) shl 8) or (longword(bytes[i*4+2]) shl 16) or (longword(bytes[i*4+3]) shl 24);
end;

procedure TDisk.write_sector(buffer : bufty);
var
  bytes: bytebuf;
  i : 0..pred(BUFSWORD);
begin
  for i := 0 to Pred(BUFSWORD) do begin
    bytes[i*4+0]:= byte((buffer[i]) and $FF);
    bytes[i*4+1]:= byte((buffer[i] shr 8) and $FF);
    bytes[i*4+2]:= byte((buffer[i] shr 16) and $FF);
    bytes[i*4+3]:= byte((buffer[i] shr 24) and $FF);
  end;
  filewrite(myfile, bytes, BUFSBYTE);
end;

end.

