unit RiscFp;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils;

type
  uint32_t = longword;
  uint8_t  = byte;
  bool     = Boolean;
  uint64_t = QWord;

  function b2i(b : boolean) : uint8_t;
  function i2b(i : uint32_t): boolean;
  function fpdiv( x,y :  uint32_t ) : uint32_t;
  function fpmul( x,y :  uint32_t) : uint32_t;
  function fpadd(x, y : uint32_t; u, v : boolean) : uint32_t;

implementation

function b2i(b : boolean) : uint8_t;
begin
  if b then b2i := 1 else b2i := 0;
end;

function i2b(i : uint32_t) : boolean;
begin
  if i <> 0 then i2b := true else i2b := false;
end;

function fpadd(x, y : uint32_t; u, v : Boolean) : uint32_t;
var
  ys, xs : Boolean;
  xm, xe, ye, ym, e0, sum, s, e1, t3, shift : uint32_t;
  y0, x0, x3, y3 : int32;

begin
  xs := (x and $80000000) <>  0;
  if not(u) then begin
    xe := (x shr 23) and $FF;
    xm := ((x and $7FFFFF) shl 1) or $1000000;
    if xs then x0 := -xm else x0 := xm;
  end else begin
    xe := 150;
    x0 := (((x and $00FFFFFF) shl 8) shr 7);
  end;

  ys := (y and $80000000) <> 0;
  ye := (y shr 23) and $FF;
  ym := ((y and $7FFFFF) shl 1);

  if (not(u) and not(v)) then ym := ym or $1000000;
  if ys then y0 := -ym else y0 := ym;

  if (ye > xe) then begin
    shift := ye - xe;
    e0 := ye;
    if (shift > 31) then x3 :=  (x0 shr 31) else x3 := (x0 shr shift);
    y3 := y0;
  end else begin
    shift := xe - ye;
    e0    := xe;
    x3    := x0;
    if (shift > 31) then
      y3 :=  (y0 shr 31)
    else
      y3 := (y0 shr shift);
  end;

   sum := ((b2i(xs) shl 26) or (b2i(xs) shl 25) or (x3 and $01FFFFFF))
    + ((b2i(ys) shl 26) or (b2i(ys) shl 25) or (y3 and $01FFFFFF));

  if ((sum and (1 shl 26)) <> 0) then s := -sum else s:= sum;
  s := (s + 1) and $07FFFFFF;

  e1 := e0 + 1;
  t3 := s shr 1;
  if ((s and $3FFFFFC) <> 0) then begin
    while ((t3 and (1 shl 24)) = 0) do begin
      t3 := t3 shl 1;
      dec(e1);
    end;
  end else begin
    t3 := t3 shl 24;
    e1 := e1 - 24;
  end;

  if v then begin
    Result := (sum shl 5) shr 6;
    exit;
  end else begin
    if ((x and $7FFFFFFF) = 0) then begin
      if not(u) then Result := y else Result := 0;
      exit;
    end;

    if ((y and $7FFFFFFF) = 0) then begin
      Result := x;
      exit;
    end;

    if ((t3 and $01FFFFFF) = 0) or ((e1 and $100) <> 0) then begin
      Result := 0;
      exit;
    end;

  end;

  Result :=  ((sum and $04000000) shl 5) or (e1 shl 23) or ((t3 shr 1) and $7FFFFF);

end;

function fpmul( x,y :  uint32_t) : uint32_t;
var
  sign, xe, ye, xm, ym, e1, z0 : uint32_t;
  m : uint64_t;
begin
  sign := (x xor y) and $80000000;
  xe   := (x shr 23) and $FF;
  ye   := (y shr 23) and $FF;

  xm    := (x and $7FFFFF) or $800000;
  ym    := (y and $7FFFFF) or $800000;
  m     := xm * ym;

  e1 := (xe + ye) - 127;

  if ((m and (1 shl 47)) <> 0) then begin
    inc(e1);
    z0 := (m shr 24) and $7FFFFF;
  end else
    z0 := (m shr 23) and $7FFFFF;

  if ((xe = 0) or (ye = 0)) then begin
    Result := 0;
    exit;
  end;

  if ((e1 and $100) = 0) then begin
    Result :=  sign or ((e1 and $FF) shl 23) or z0;
    exit;
  end;

  if ((e1 and $80) = 0) then begin
    Result := sign or ($FF shl 23) or z0;
    exit;
  end;

  Result := 0;

end;


function fpdiv( x,y :  uint32_t ) : uint32_t;
var
  e1, sign, xe, ye,
  xm, ym, q1,  q2 : uint32_t;
begin
  sign := (x xor y)  and $80000000;
  xe   := (x shr 23) and $FF;
  ye   := (y shr 23) and $FF;
  xm   := (x and $7FFFFF) or $800000;
  ym   := (y and $7FFFFF) or $800000;
  q1   := (xm * (1 shl 23) DIV ym);
  e1   := (xe - ye) + 126;

  if ((q1 and $800000) <> 0) then begin
    inc(e1);
    q2 := q1 and $7FFFFF;
  end else
    q2 := (q1 shl 1) and $7FFFFF;

  if (xe = 0) then begin
    Result := 0;
    exit;
  end;

  if (ye = 0) then begin
    Result :=  sign or ($FF shl 23);
    exit;
  end;

  if ((e1 and $100) = 0) then begin
    Result := sign or ((e1 and $FF) shl 23) or q2;
    exit;
  end;

  if ((e1 and $80) = 0) then begin
    Result :=  sign or ($FF shl 23) or q2;
    exit;
  end;

  Result := 0;

end;

end.

