unit RiscMain;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, LclIntf, Controls, Forms,
  BGRABitmap, BGRABitmapTypes,
  Machines, RiscCpu, RiscFp, RiscKeyCode,

  LcLType, Graphics;

type

  { TRisc }

  { TRiscCore }

  TRiscCore = class(TMachine)
  private
    Risc  : TRisc;
    FImage: TBGRABitmap;
  public
    constructor Create(Sender: TObject; filename : string);
    destructor  Destroy; override;
    procedure   IdleProc(Sender: TObject); override;
    procedure   KeyDown(Sender: TObject; Key: Word; Shift: TShiftState); override;
    procedure   MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure   MouseMove(Sender: TObject; Shift: TShiftState; X,Y: Integer); override;
    procedure   MouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure   Paint(Sender: TObject); override;
  end;

implementation

var
  Hash : array [0..Pred(MemWords)] of uint32_t;

const
  FPGABlack: TBGRAPixel = (blue: $83; green: $7B; red: $65; alpha: 255);
  FPGAWhite: TBGRAPixel = (blue: $E3; green: $F6; red: $FD; alpha: 255);

{ TRisc }

constructor TRiscCore.Create(Sender: TObject; filename: string);
begin
  Risc.Create(filename);
  TForm(Sender).Width:=RISC_SCREEN_WIDTH;
  TForm(Sender).Height:=RISC_SCREEN_HEIGHT;
  FImage := TBGRABitmap.Create(RISC_SCREEN_WIDTH,RISC_SCREEN_HEIGHT);
  FImage.FillRect(0,0,RISC_SCREEN_WIDTH,RISC_SCREEN_HEIGHT,FPGABlack,dmSet);
end;

destructor TRiscCore.Destroy;
begin
  Risc.Destroy;
  FImage.free;
end;

procedure TRiscCore.IdleProc(Sender: TObject);
var
  frame_start: uint32_t;
  delay: longint;
begin
  frame_start:=GetTickCount;
  Risc.set_time(frame_start);
  Risc.run(CPU_HZ div FPS);
  delay := frame_start + (1000 div FPS) - GetTickCount;
  if delay > 0 then Sleep(delay);
  LclIntf.InvalidateRect(TForm(Sender).Handle,NIL,False);
end;

procedure TRiscCore.KeyDown(Sender: TObject; Key: Word; Shift: TShiftState);
var
  Buff: keybuf;
  Len : uint32_t;
begin
  Len:=RiscKeyDown(Key,Shift,Buff);
  Risc.keyboard_input(Buff,Len);
end;

procedure TRiscCore.MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  b: integer;
begin
  case Button of
    mbLeft   : b:=1;
    mbRight  : b:=3;
    mbMiddle : b:=2;
  end;
  Risc.mouse_button(b,true);
end;

procedure TRiscCore.MouseMove(Sender: TObject; Shift: TShiftState; X,Y: Integer);
var
  XX,YY: integer;
begin
  XX:=X*RISC_SCREEN_WIDTH div TForm(Sender).Width;
  YY:=(TForm(Sender).Height-Y)*RISC_SCREEN_HEIGHT div TForm(Sender).Height;
  Risc.mouse_moved(XX,YY);
end;

procedure TRiscCore.MouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  b: integer;
begin
  case Button of
    mbLeft   : b:=1;
    mbRight  : b:=3;
    mbMiddle : b:=2;
  end;
  Risc.mouse_button(b,false);
end;

procedure TRiscCore.Paint(Sender: TObject);
var
  idx   : integer;
  pixels: uint32_t;
  tmppix: uint32_t;
  Y,X   : integer;
  b     : integer;
begin

  idx := Displaystart div 4;

  for Y :=RISC_SCREEN_HEIGHT-1 downto 0 do
    for X := 0 to RISC_SCREEN_WIDTH div 32 -1 do begin
       pixels := risc.RAM[idx];
       tmppix := Hash[idx];
       if Hash[idx]<>pixels then
         for b := 0 to 31 do begin
           if (pixels and 1) <> (tmppix and 1) then
             if ((pixels and 1) > 0) then
               FImage.DrawPixel(X*32+B,Y,FPGAWhite)
             else
               FImage.DrawPixel(X*32+B,Y,FPGABlack);
           pixels := pixels shr 1;
           tmppix := tmppix shr 1;
         end;
       Hash[idx] := risc.RAM[idx];
       inc(idx);
    end;

  //FImage.InvalidateBitmap;
  FImage.Draw(TForm(Sender).Canvas,Classes.Rect(0,0,TForm(Sender).Width,TForm(Sender).Height),False);

end;

(*
function GetDIBPointer(BitMap: TBitmap):Pointer;
var
  BInfo: TBitmapInfo;
begin
  BInfo.bmiHeader.biSize := sizeof(BitmapInfo.bmiHeader);
  BInfo.bmiHeader.biWidth := BitMap.Width;
  BInfo.bmiHeader.biHeight := -BitMap.Height;
  BInfo.bmiHeader.biPlanes := 1;
  BInfo.bmiHeader.biBitCount := 32;
  BInfo.bmiHeader.biCompression := BI_RGB;
  BitMap.Handle := CreateDIBSection(BitMap.Canvas.Handle, BInfo, DIB_RGB_COLORS, GetDIBPointer, 0, 0);
end;

procedure TRiscCore.Paint(Sender: TObject);
type
  TBig = array[0..0] of Integer;
var
  BitMap : TBitMap;
  bmpPointer : Pointer; //для быстрого доступа к пикселям битмапа
  idx   : integer;
  pixels: uint32_t;
  tmppix: uint32_t;
  Y,X   : integer;
  b     : integer;
begin
  BitMap:= Graphics.TBitmap.Create;
  BitMap.Width  := TForm(Sender).Width;
  BitMap.Height := TForm(Sender).Height;
  bmpPointer := GetDIBPointer(BitMap);

  idx := Displaystart div 4;

  for Y :=RISC_SCREEN_HEIGHT-1 downto 0 do
    for X := 0 to RISC_SCREEN_WIDTH div 32 -1 do begin
       pixels := risc.RAM[idx];
         for b := 0 to 31 do begin
           if (pixels and 1) > 0 then
             TBig(bmpPointer^)[x*32 + b + y * BitMap.width] := $FFFFFF;
           pixels := pixels shr 1;
         end;
       inc(idx);
    end;

  TForm(Sender).Canvas.StretchDraw(Classes.Rect(0,0,BitMap.Width,BitMap.Height),BitMap);
  BitMap.Free;
end;
*)

end.

